package sorting;

public class MergeSortArray implements SortingAlgorithm<Integer>
{
    /**
     * Mergesort algorithm for an array.
     * @param a an array of Comparable items to sort.
     * @return sorting statistics.
     */
    @Override
    public Stats sort(Integer[] a) 
    {
        Integer[] tmpArray = new Integer[a.length];
        return mergeSort(a, tmpArray, 0, a.length - 1);
    }

    /**
     * Internal method for mergesort that makes recursive calls.
     * @param a an array of Comparable items.
     * @param tmpArray an array to place the merged result.
     * @param left the left-most index of the subarray.
     * @param right the right-most index of the subarray.
     * @return sorting statistics.
     */
    private Stats mergeSort(Integer[] a, Integer[] tmpArray, int left, int right) 
    {
        Stats stats1 = new Stats();
        Stats stats2 = new Stats();
        Stats stats3 = new Stats();
        
        if (left < right) {
            int center = (left + right)/2;
            
            stats1 = mergeSort(a, tmpArray, left, center);
            stats2 = mergeSort(a, tmpArray, center+1, right);
            stats3 = merge(a, tmpArray, left, center+1, right);
        }
        
        return new Stats(stats1.moves    + stats2.moves    + stats3.moves,
                         stats1.compares + stats2.compares + stats3.compares);
    }

    /**
     * Internal method for mergesort that merges two sorted subarray halves.
     * @param a an array of Comparable items.
     * @param tmpArray an array to place the merged result.
     * @param leftPos the left-most index of the subarray.
     * @param rightPos the index of the start of the second half.
     * @param rightEnd the right-most index of the subarray.
     * @return sorting statistics.
     */
    private Stats merge(Integer[] a, Integer[] tmpArray, 
                        int leftPos, int rightPos, int rightEnd) 
    {
        int leftEnd = rightPos - 1;
        int tmpPos = leftPos;
        int numElements = rightEnd - leftPos + 1;
        long moves = 0;
        long compares = 0;

        // Main loop
        while (leftPos <= leftEnd && rightPos <= rightEnd) {
            if (a[leftPos].compareTo(a[rightPos]) <= 0) {
                tmpArray[tmpPos++] = a[leftPos++];
            }
            else {
                tmpArray[tmpPos++] = a[rightPos++];
            }
            moves++;
            compares++;
        }

        // Copy rest of first half.
        while (leftPos <= leftEnd) {
            tmpArray[tmpPos++] = a[leftPos++];
            moves++;
        }

        // Copy rest of right half.
        while (rightPos <= rightEnd) {
            tmpArray[tmpPos++] = a[rightPos++];
            moves++;
        }

        // Copy tmpArray back.
        for (int i = 0; i < numElements; i++, rightEnd--) {
            a[rightEnd] = tmpArray[rightEnd];
            moves++;
        }
        
        return new Stats(moves, compares);
    }

    @Override
    public Stats sort(MyLinkedList<Integer> list) { return null; };
}
