package sorting;

public class ShellSortKnuth implements SortingAlgorithm<Integer>
{
    /**
     * Shellsort using Knuth's increment sequence for h.
     * @param a an array of Comparable items to sort.
     * @return sorting statistics.
     */
    @Override
    public Stats sort(Integer[] a) 
    {
        int j;
        long totalMoves = 0;
        long totalCompares = 0;

        // Compute highest h value.
        int h = 1;
        while (h < a.length) h = 3*h + 1; 
        
        for (h = (h-1)/3; h >= 1; h = (h-1)/3) {            
            for (int pass = h; pass < a.length; pass++) {
                long moves = 0;
                long compares = 0;
                Integer tmp = a[pass];
                
                for (j = pass; j >= h && tmp.compareTo(a[j-h]) < 0; j -= h) {
                    a[j] = a[j-h];
                    moves++;
                    compares++;
                }
                
                if (j >= h) compares++;
                
                if (pass != j) {
                    a[j] = tmp;
                    moves++;
                }
                
                totalMoves += moves;
                totalCompares += compares;
            }
        }
        
        return new Stats(totalMoves, totalCompares);
    }

    @Override
    public Stats sort(MyLinkedList<Integer> list) { return null; };
}
